#!/usr/bin/env python3
import matplotlib.pyplot as plt
import pandas as pd

week = pd.Timedelta('7d')
def weekly_plot(dataframe, datacol='x', **kwargs):
    groups = dataframe.groupby([dataframe.index.year, # dataframe.index.month,
                                dataframe.index.week])
    n = len(groups)
    fig, axes = plt.subplots(n, 1, sharey=True, figsize=(8, 2*n))
    for i, group in enumerate(groups):
        ax = axes[i] if str(type(axes)) == "<class 'numpy.ndarray'>" else axes
        start = pd.Timestamp(str(group[0][0])) + pd.Timedelta((group[0][1]-1)*7, 'd')
        end = start + week
        group[1].plot(y=datacol, ax=ax, **kwargs)
        ax.set_xlim(start, end)
        if ax.legend_:
            ax.legend_.remove()
    plt.subplots_adjust(hspace=0.5)
    fig.tight_layout()
    return fig

day = pd.Timedelta('1d')
def daily_plot(dataframe, datacol, **kwargs):
    groups = dataframe.groupby([dataframe.index.year, dataframe.index.month,
                                dataframe.index.day])
    n = len(groups)
    fig, axes = plt.subplots(n, 1, sharey=True, figsize=(3, 0.15*n))
    for i, group in enumerate(groups):
        ax = axes[i] if str(type(axes)) == "<class 'numpy.ndarray'>" else axes
        start = pd.Timestamp(group[1].index[0].date())
        end = start + day
        group[1].plot(y=datacol, ax=ax, legend=False, **kwargs)
        ax.set_xlim(start, end)
        ax.set_yticks([])
        ax.set_xticks([])
        ax.set_xlabel('')
        ax.set_ylabel('')
        plt.axis('off')
        if hasattr(ax, 'legend_') and ax.legend_:
            ax.legend_.remove()
    plt.subplots_adjust(hspace=0)
    plt.tick_params(axis='both', which='both', bottom=False, top=False,
            labelbottom=False, right=False, left=False, labelleft=False)
    return fig



""" Validator functions for RADAR data dataframes """
import pandas as pd


def android_phone_contacts(df: pd.DataFrame) -> pd.Series:
    """ Are rows in a phone_contacts dataframe valid
    There are occasions where there are a large number of contacts listed
    (100,000+), which are clearly erroneous. Additionally, sometimes all
    contacts are apparently removed and then readded near instantaneously.
    This function returns false for rows with these problems.
    Params:
        df (pd.DataFrame): Dataframe with 'contacts', 'contactsAdded',
            and 'contactsRemoved' columns.
    Returns:
        pd.Series[bool]
    """
    valid = (
        (df['contacts'] < 32000) &
        (df['contactsAdded'] < (1+df['contacts']//2)) &
        (df['contactsRemoved'] < (1+df['contacts']//2))
    )
    return valid


def questionnaire_phq8(df: pd.DataFrame) -> pd.Series:
    """ Validates PHQ8 responses
    Ensures there are 8 answers with values 0 <= v <= 3
    Params:
        df (pd.DataFrame): Melted PHQ8 dataframe
            (questionnaire dataframes are melted on load by default)
    Returns:
        pd.Series[bool]
    """
    valid = ((df['answers_value'].astype(int) <= 3) &
             (df['answers_value'].astype(int) >= 0))
    valid.loc[df.groupby(df.index).apply(len) != 8] = False
    return valid


def questionnaire_rses(df: pd.DataFrame) -> pd.Series:
    """ Validates RSES responses
    Ensures there are 10 answers with values 0 <= v <= 3
    Params:
        df (pd.DataFrame): Melted RSES dataframe
            (questionnaire dataframes are melted on load by default)
    Returns:
        pd.Series[bool]

    """
    valid = ((df['answers_value'].astype(int) <= 3) &
             (df['answers_value'].astype(int) >= 0))
    valid.loc[df.groupby(df.index).apply(len) != 10] = False
    return valid


VALIDATORS = {
    'android_phone_contacts': android_phone_contacts,
    'questionnaire_phq8': questionnaire_phq8,
    'questionnaire_rses': questionnaire_rses
}



import pandas as pd
import seaborn as sns
from datetime import datetime
import matplotlib.patches as patches

from ..common import log
from ..util.completion import completion_idx_has_data

def completion_plot(completion, modalities, start, end, freq,
    ax=None, cmap=None, x_tick_mult=24, x_tick_fmt="%y-%m-%d %H:%M",
    enrolment_date=None, events=None, event_radius=0):
    """ Plots the completion array as a heatmap
    Parameters
    __________
    completion: np.ndarray
        Completion 2d array for a participant
    modalities: list of str
        A list of the modalities, used to label the Y-axis
    start: pd.Timedelta / np.datetime64
    end: pd.Timedelta / np.datetime64
    freq: str
        The timedelta string associated with the given completion array
    ax: matplotlib axes
    cmap: Colormap
    x_tick_mult: int
        Multiplier for x ticks, will draw a tick every x_tick_mult hours.
    x_tick_fmt: str
    enrolment_date: pd.Timedelta / np.datetime64
    events: list of tuples of strings
        A list of event timestamp and label pairs, both as strings
        Draws a vertical line at each event, colored as:
            green: data from all modalities is present at the event
            orange: data from at least one modality is present at the event
            red: no data is present at the event
        If event_radius is specified the color checks include the whole event circumference.
    event_radius: float
        Radius around each event in multiples of freq

    Returns
    _______
    ax: matplotlib axes
    """
    td = pd.Timedelta(freq)
    ax = sns.heatmap(completion, cmap=cmap, cbar=False, xticklabels=int(x_tick_mult*(pd.Timedelta('1h')/td)), ax=None)
    # Gaps
    N_x = completion.shape[1]
    N_y = completion.shape[0]
    for i in range(N_y-1):
        ax.hlines(i+1, 0, N_x, color='white', linewidth=2)

    # Outline
    ax.vlines(0, 0, N_y, color='black', linewidth=2)
    ax.vlines(N_x, 0, N_y, color='black', linewidth=2)
    ax.hlines(0, 0, N_x, color='black', linewidth=2)
    ax.hlines(N_y, 0, N_x, color='black', linewidth=2)

    # Events
    if events:
        for e_stamp_str,e_label in events:
            try:
                e_stamp_unix = float(e_stamp_str)
                e_stamp = pd.Timestamp(e_stamp_unix, unit='s').tz_localize('UTC')
            except:
                # TODO: review usage of tz_localize()
                e_stamp = pd.Timestamp(datetime.strptime(e_stamp_str, '%Y-%m-%d %H:%M:%S')).tz_localize('CET').tz_convert('UTC')

            if e_stamp < start or e_stamp > end: continue

            log.debug("Event at {}: {}".format(e_stamp, e_label))
            e_idx = (e_stamp - start)//td

            e_slice = None
            if event_radius:
                e_start = max(0, (e_stamp - start - (td * event_radius))//td)
                e_end = min(N_x-1, (e_stamp - start + (td * event_radius))//td)
                e_slice = slice(e_start, e_end+1)
                rect = patches.Rectangle((e_start, 0), e_end - e_start, N_y, linewidth=0.5, edgecolor='k', alpha=0.25, zorder=9)
                ax.add_patch(rect)

            has_all = completion_idx_has_data(completion, e_slice if e_slice else e_idx, requirement_function=all)
            has_any = completion_idx_has_data(completion, e_slice if e_slice else e_idx, requirement_function=any)

            if has_all: ax.vlines(e_idx, 0, N_y, color='green', linewidth=2, zorder=10)
            elif has_any: ax.vlines(e_idx, 0, N_y, color='orange', linewidth=2, zorder=10)
            else: ax.vlines(e_idx, 0, N_y, color='red', linewidth=2, zorder=10)

    # Enrolment
    if enrolment_date:
        enrolment_idx = (enrolment_date - start)//td
        ax.vline(enrolment_idx-1, 0, N_y, color='red', linewidth=0.5)

    # Labels
    ax.set_ylabel('Data topic')
    ax.set_xlabel('Date')
    xticks = ax.get_xticks()
    ax.set_xticklabels([(start + (tm * td)).strftime(x_tick_fmt) for tm in xticks], rotation=0)
    ax.set_yticklabels([ m.split('_', 2)[-1] for m in modalities ], rotation=0)

    return ax



import os
import matplotlib.pyplot as plt
import pandas as pd
import dask.dataframe as dd

from ..common import log
from ..util.parsers import timestamp_from_string


pd.plotting.register_matplotlib_converters()


def data_plot(ptc, modalities, start, end, freq, ax=None, events=None, event_bounds=None, event_resample=None, outdir=None):
    """ Plots the given participant data as lineplots
    Parameters
    __________
    ptc: Participant
    modalities: list of str
        A list of the modalities to plot
    start: pd.Timedelta / np.datetime64
    end: pd.Timedelta / np.datetime64
    freq: str
        The data is resampled at the specified frequency
    ax: matplotlib axes
    events: list of tuples
        A list of event timestamp and boolean pairs, a detailed plot of each event with boolean=True will be created, see event_range
    event_bounds: list of tuples
        A list of Timedelta pairs, if specified detailed plots of events will be made for each entry in the list, with the respective bounds
    event_resample: str
        If specified, each event detail plot is resampled at the specified frequency, instead of plotting the raw data
    outdir: str
        If specified, the figure will be saved as an image in that directory, instead of shown as a pyplot window
    """
    if modalities is None:
        modalities = sorted(list(ptc.data.keys()))

    fig = plt.figure(figsize=(20,10), num="{} - raw data".format(ptc.name))

    for i, dname in enumerate(modalities):
        if dname not in ptc.data or not isinstance(ptc.data[dname], dd.DataFrame):
            log.debug("Participant %s has no recorded %s", ptc.name, dname)
            continue
        log.info("Plotting %s for participant %s", dname, ptc.name)
        ddf = ptc.data[dname]
        resampled = ddf.resample(freq).mean()
        resampled = resampled.compute()
        ax = fig.add_subplot(len(modalities), 1, i+1, sharex=ax)
        ax.plot(resampled)
        ax.autoscale(enable=True, tight=True)
        ax.set_title("{} - {}".format(ptc.name, dname.split('_', 2)[-1]))

    log.info("Plotting %d events for participant %s", len(events) if events else 0, ptc.name)
    for ev in events:
        e = timestamp_from_string(ev[0])
        if e < start or e > end: continue

        for subax in fig.get_axes(): subax.axvline(e, color='red', linewidth=2, zorder=10)
        for b in event_bounds:
            if ev[1]: data_detail_plot(ptc, e, modalities, b, outdir, event_resample)

    if not outdir: fig.set_tight_layout(True)

    if outdir: fig.savefig('{}{}{}_data.png'.format(outdir, os.path.sep, ptc.name), dpi=288, bbox_inches='tight')
    else: plt.show()
    plt.close(fig)


def data_detail_plot(ptc, ev, modalities, bounds, outdir=None, resample=None):
    """ Plots the given event in detail as lineplots using bounds as the start and end of the plot, relative to ev
    Parameters
    __________
    ptc: Participant
    ev: datetime
        The event timestamp
    modalities: list of str
        A list of the modalities to plot
    bounds: tuple
        A Timedelta pair, specifying the bounds before and after the event to plot
    resample: str
        If specified, the data is resampled at the specified frequency, instead of plotting the raw data
    outdir: str
        If specified, the figure will be saved as an image in that directory, instead of shown as a pyplot window
    """
    ev_fig = plt.figure(figsize=(20,10), num="{} - event at {} - {} - {}-{}".format(ptc.name, ev, resample if resample else 'raw', int(bounds[0].seconds/60), int(bounds[1].seconds/60)))
    ev_pre = ev - bounds[0]
    ev_post = ev + bounds[1]
    ev_ax=None

    for i, dname in enumerate(modalities):
        if dname not in ptc.data or not isinstance(ptc.data[dname], dd.DataFrame): continue
        ddf = ptc.data[dname]
        if resample: ev_data = ddf.resample(resample).mean().loc[ev_pre:ev_post].dropna().compute()
        else: ev_data = ddf.loc[ev_pre:ev_post] \
            .drop('projectId',axis=1,errors='ignore') \
            .drop('userId',axis=1,errors='ignore') \
            .drop('sourceId',axis=1,errors='ignore') \
            .drop('timeReceived',axis=1,errors='ignore') \
            .dropna().compute()
        ev_ax = ev_fig.add_subplot(len(modalities), 1, i+1, sharex=ev_ax)
        ev_ax.plot(ev_data)
        ev_ax.axvline(ev, color='red', linewidth=2, zorder=10)
        ev_ax.autoscale(enable=True, tight=True)
        ev_ax.set_title("{} - {}".format(ptc.name, dname.split('_', 2)[-1]))

    if not outdir: ev_fig.set_tight_layout(True)

    if outdir: ev_fig.savefig('{}{}{}_event_{}_{}_{}-{}.png'.format(outdir, os.path.sep, ptc.name, ev.strftime('%Y%m%d-%H%M%S'), resample if resample else 'raw', int(bounds[0].seconds/60), int(bounds[1].seconds/60)), dpi=288, bbox_inches='tight')
    else: plt.show()
    plt.close(ev_fig)



#!/usr/bin/env python3
import requests
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient


class Resource():
    """ A generic resource object to form part of the
    ManagementPortal object. Sends requests using the
    ManagementPortal.request method
    """
    _name = ''

    def __init__(self, portal):
        self._portal = portal

    def request(self, method, endpoint='', **kwargs):
        endpoint = self._name + '/' + endpoint if endpoint else self._name
        return self._portal.request(method, endpoint, **kwargs)


class Subject(Resource):
    _name = 'subjects'

    def get(self, login='', **params):
        return self.request('GET', endpoint=login,
                            params=params)

    def get_all(self, **params):
        return self.get('', **params)

    def create(self, **body):
        return self.request('POST', json=body)

    def update(self, **body):
        return self.request('PUT', json=body)

    def discontinue(self, **body):
        return self.request('PUT', endpoint='discontinue', json=body)

    def delete(self, login):
        return self.request('DELETE', login)

    def get_sources(self, login):
        return self.request('GET', login + '/sources')

    def assign_sources(self, login, **body):
        if not (('sourceTypeId' in body) or
                any(x in body for x in ('sourceTypeProducer',
                                        'sourceTypeModel',
                                        'sourceTypeCatalogVersion'))):
            raise ValueError(('You must supply either a Source Type ID, or the'
                              'combination of (sourceTypeProducer,'
                              'sourceTypeModel, sourceTypeCatalogVersion)'))
        return self.request('POST', endpoint=login, json=body)

    def revisions(self, login):
        return self.request('GET', endpoint=login + '/revisions')


class Project(Resource):
    _name = 'projects'

    def get(self, project_name='', **params):
        return self.request('GET', endpoint=project_name,
                            params=params if not project_name else None)

    def get_all(self, **params):
        return self.get('', **params)

    def create(self, project_name, description, location, **body):
        body['projectName'] = project_name
        body['description'] = description
        body['location'] = location
        return self.request('POST', json=body)

    def update(self, project_name, description, location, **body):
        body['projectName'] = project_name
        body['description'] = description
        body['location'] = location
        return self.request('PUT', json=body)

    def delete(self, project_name):
        return self.request('DELETE', project_name)

    def get_roles(self, project_name):
        endpoint = project_name + '/roles'
        return self.request('GET', endpoint)

    def get_source_types(self, project_name):
        endpoint = project_name + '/source-types'
        return self.request('GET', endpoint)

    def get_sources(self, project_name, **params):
        endpoint = project_name + '/sources'
        return self.request('GET', endpoint, params=params)

    def get_subjects(self, project_name, **params):
        endpoint = project_name + '/subjects'
        return self.request('GET', endpoint, params=params)


class Source(Resource):
    _name = 'sources'

    def get(self, source_name='', **params):
        return self.request('GET', endpoint=source_name,
                            params=params if not source_name else None)

    def get_all(self, **params):
        return self.get('', **params)

    def create(self, source_name, source_type, assigned, **body):
        body['assigned'] = assigned
        body['sourceName'] = source_name
        body['sourceType'] = source_type
        return self.request('POST', json=body)

    def update(self, source_name, source_type, assigned, **body):
        body['assigned'] = assigned
        body['sourceName'] = source_name
        body['sourceType'] = source_type
        return self.request('PUT', json=body)

    def delete(self, source_name):
        return self.request('DELETE', source_name)


class SourceType(Resource):
    _name = 'source-types'

    def get(self, producer='', model='', version='', **params):
        if model and not producer:
            raise ValueError('You must specify a producer to specify a model')
        if version and not model:
            raise ValueError('You must specify a model to specify a version')
        endpoint = '/'.join([x for x in (producer, model, version) if x])
        return self.request('GET', endpoint=endpoint,
                            params=params if not model else None)

    def get_all(self, **params):
        return self.get(**params)

    def create(self, producer, model, version, can_register_dynamically, scope,
               **body):
        body['producer'] = producer
        body['model'] = model
        body['catalogVersion'] = version
        body['canRegisterDynamically'] = can_register_dynamically
        body['sourceTypeScope'] = scope
        return self.request('POST', json=body)

    def update(self, producer, model, version, can_register_dynamically, scope,
               **body):
        body['producer'] = producer
        body['model'] = model
        body['catalogVersion'] = version
        body['canRegisterDynamically'] = can_register_dynamically
        body['sourceTypeScope'] = scope
        return self.request('PUT', json=body)

    def delete(self, producer, model, version):
        endpoint = '/'.join([producer, model, version])
        return self.request('DELETE', endpoint)


class SourceData(Resource):
    _name = 'source-data'

    def get(self, source_data_name='', **params):
        return self.request('GET', source_data_name,
                            params=params if not source_data_name else None)

    def get_all(self, **params):
        return self.get(**params)

    def create(self, source_data_type, **body):
        body['sourceDataType'] = source_data_type
        return self.request('POST', json=body)

    def update(self, source_data_type, **body):
        body['sourceDataType'] = source_data_type
        return self.request('PUT', json=body)

    def delete(self, source_data_name):
        return self.request('DELETE', source_data_name)


class ManagementPortal():
    """ A class to call resource REST API requests from. Requires
    a RADAR management portal API account.
    """

    def __init__(self, client_id, client_secret, url):
        self._url = url.rstrip('/')
        self.s = requests.Session()
        self.s.headers['Authorization'] = \
            'Bearer {}'.format(self.authenticate(client_id, client_secret))
        self.subjects = Subject(portal=self)
        self.projects = Project(portal=self)
        self.source = Source(portal=self)
        self.source_type = SourceType(portal=self)
        self.source_data = SourceData(portal=self)

    def authenticate(self, client_id, client_secret):
        def get_token(client_id, client_secret, oauth, url):
            token_url = url + '/oauth/token'
            token = oauth.fetch_token(token_url=token_url,
                                      client_id=client_id,
                                      client_secret=client_secret)
            return token
        client = BackendApplicationClient(client_id=client_id)
        oauth = OAuth2Session(client=client)
        token = get_token(client_id, client_secret, oauth, self._url)
        return token['access_token']

    def request(self, method, endpoint, **kwargs):
        url = self._url + '/api/' + endpoint
        r = self.s.request(method, url, **kwargs)
        if (r.status_code == 401 and 'error_description' in r.json() and
                'Access token expired' in r.json()['error_description']):
            err = ('401 Client Error: Access token has expired')
            raise requests.HTTPError(err)
        r.raise_for_status()
        return r



""" Configuration defaults and classes
Contains the config and also default schemas, specifications,
protocols, and definitions.
"""
import os
import collections
import yaml

schemas = {}
specifications = {}
protocols = {}
definitions = {}

_ROOT = os.path.abspath(os.path.dirname(__file__))
_DEF_RADAR_SCHEMA_PATH = os.path.join(_ROOT, 'Schemas', 'commons')
_DEF_RADAR_SPECS_PATH = os.path.join(_ROOT, 'Schemas', 'specifications')


class Config(dict):
    """ Class that allows access to values through getattr,
    and converts the class of setattr/setitem dicts to Config
    """
    def __getitem__(self, key):
        return dict.get(self, key, None)

    def __getattr__(self, key):
        return self.get(key, None)

    def __setattr__(self, key, val):
        self[key] = val

    def __setitem__(self, key, val):
        if isinstance(val, dict):
            val = Config(val)
        dict.__setitem__(self, key, val)


config = Config()
config['aRMT'] = Config()
config['io'] = Config()
config['log'] = Config()
config['log']['to_file'] = None
config['log']['format'] = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
config['protocol'] = Config()
config['schema'] = Config()
config['specifications'] = Config()
config['project'] = Config()
config['schema']['device'] = 'android'
config['schema']['key'] = None
config['schema']['dir'] = _DEF_RADAR_SCHEMA_PATH
config['schema']['git'] = False
config['schema']['github_owner'] = 'RADAR-base'
config['schema']['github_repo'] = 'RADAR-Schemas'
config['schema']['github_sha'] = None
config['schema']['read_csvs'] = True
config['schema']['from_local_file'] = False
config['schema']['local_file_regex'] = '.*schema.*json'
config['specifications']['dir'] = _DEF_RADAR_SPECS_PATH
config['specifications']['git'] = False
config['specifications']['github_owner'] = 'RADAR-base'
config['specifications']['github_repo'] = 'RADAR-Schemas'
config['specifications']['github_sha'] = 'extended-specs'
config['io']['index'] = 'time'
config['protocol']['url'] = ''
config['project']['ptckw'] = {}
config['project']['datakw'] = {}


def update(d, u):
    """ update nested mappings
    d (dict): mapping to update
    u (dict): mapping to update with
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


_CONFIG_DEFAULT_FILE = os.path.join(_ROOT, 'config.yml')
if os.path.exists(_CONFIG_DEFAULT_FILE):
    with open(_CONFIG_DEFAULT_FILE) as f:
        update(config, yaml.load(f, Loader=yaml.FullLoader))

if 'config.yml' in os.listdir():
    with open('config.yml', 'r') as f:
        update(config, yaml.load(f, Loader=yaml.FullLoader))

if config.schema.dir:
    from radar.util.schemas import schemas_from_commons
    schemas.update(schemas_from_commons(config.schema.dir))

if config.specifications.dir:
    from radar.util.specifications import specifications_from_directory
    specifications.update(specifications_from_directory())
elif config.specifications.git:
    from radar.util.specifications import specifications_from_github
    specifications.update(specifications_from_github())



#!/usr/bin/env python3
import re
import numpy as np
import pandas as pd
from ..common import log

_DEF_COLS = ['key.projectId', 'key.userId', 'key.sourceId',
             'value.time', 'value.timeCompleted', 'value.name',
             'value.version']


def iterrows(df):
    return (row[1] for row in df.iterrows())


def populate(df, definition, fields=None):
    """ Populates a melted aRMT dataframe with additional columns based
    on the questionId
    Parameters
    __________
    df : DataFrame
        A dataframe containing aRMT data as read from an extracted RADAR csv.
    definition : list of dicts
        An aRMT definition loaded with json std lib.
    fields : dict
        A dict with field names as keys and functions as values. The function
        should have the definition entry as first argument and the code/answer
        value as the second argument.

    Returns
    ______
    DataFrame
        A DataFrame with the additional column fields.

    """

    def add_columns(row):
        """ Adds given columns to the row using the definition.
        Parameters
        _________
        row: pandas dataframe row
            A row corresponding to a single questionnaire submission.
        definition: list of dicts
            The corresponding RADAR definition parsed with the json std library
        Returns
        _______
        row_dict: dict
            A dict representation of the row from which a dataframe can be made
        """
        out_row = row.to_dict()
        for i in range(n):
            qid = out_row['value.answers.{}.questionId'.format(i)]
            entry = defintion_entry_from_label(definition, qid)
            code = row['value.answers.{}.value'.format(i)]
            for field, func in fields.items():
                key = 'value.answers.{}.{}'.format(i, field)
                out_row[key] = func(entry, code)
        return out_row

    if fields is None:
        fields = {'questionLabel': key_func('field_label'),
                  'valueLabel': code_label}
    n = 1 + max([int(c.split('.')[2]) for c in df if 'answers' == c[6:13]])
    out = [add_columns(row) for row in iterrows(df)]
    return pd.DataFrame(out, columns=arrange_cols(fields, n))


def code_label(entry, code):
    choices = entry['select_choices_or_calculations']
    labels = [c['label'] for c in choices if
              c['code'].strip() == str(code).strip()]
    if len(labels) != 1:
        err = ('Can not map answer code "{}"'
               ' to the following possibilities in question "{}":\n'
               '{}').format(code, entry['field_name'], choices)
        log.warning(err)
        return np.nan
    return labels[0]


def key_func(key):
    def use_key(entry, *args, **kwargs):
        return entry[key]
    return use_key


def arrange_cols(fields, n):
    cols = _DEF_COLS.copy()
    for i in range(n):
        cols.extend('value.answers.{}.{}'.format(i, f)
                    for f in sorted(fields))
        cols.extend('value.answers.{}.{}'.format(i, c)
                    for c in ('questionId', 'value', 'startTime', 'endTime')
                    if c not in fields)
    return cols


def defintion_entry_from_label(definition, label):
    for entry in definition:
        if entry['field_name'] == label:
            return entry
    raise ValueError('No such label {} in definition'.format(label))



#!/usr/bin/env python3
import numpy as np


def biovotion_sort_upsample(df, time_idx='value.time'):
    df = df.sort_index(ascending=False).sort_values(by=time_idx, kind='mergesort')
    lens = df.groupby('value.time').apply(len).values
    deltas = np.array([i/j for j in lens for i in range(j)])
    df[time_idx] = (df[time_idx].astype(int) +
                    deltas.dot(10**9).astype(int)).astype(np.datetime64)
    return df


def timestamp_from_string(stampstr, fromtz='CET', totz='UTC'):
    import pandas as pd
    from datetime import datetime
    try:
        sz_unix = float(stampstr)
        return pd.Timestamp(sz_unix, unit='s').tz_localize('UTC').tz_convert(totz)
    except:
        return pd.Timestamp(datetime.strptime(stampstr, '%Y-%m-%d %H:%M:%S')).tz_localize(fromtz).tz_convert(totz)



import numpy as np
import pandas as pd
import seaborn as sns

HOUR = pd.Timedelta('1h')
CMAP = [(1, 1, 1)]
CMAP.extend(sns.color_palette('Set2', 20))


def ptc_number(ptc):
    return int(ptc.info['record_id'])


def data_completion_hourly(ddf, start, end, enrolment_date=None):
    """ Returns a vector with elements representing whether data is
    present at an hourly resolution.
    Parameters
    __________
    ddf: dask.dataframe
        A delayed dataframe as produced through radar.io.load().
        Must have a DatetimeIndex.
    start: pd.Timestamp / np.datetime64
    end: pd.Timestamp / np.datetime64
    enrolment_date: pd.Timestamp / np.datetime64 (optional)
        Hours before this date are assigned -1.

    Returns
    _______
    v: np.ndarray
        Array where 1 corresponds to present data, 0 to missing data,
        and -1 to prior to enrolment.
    """
    vec = pd.Series(index=pd.date_range(start=start, end=end, freq='1h'),
                    dtype=int)
    idx = pd.DatetimeIndex(ddf.divisions)
    idx = idx[np.logical_and(idx > vec.index[0], idx < vec.index[-1])]
    vec.loc[idx] = 1
    if enrolment_date:
        vec.loc[:enrolment_date] = -1
    return vec.values


def ptc_completion_hourly(ptc, start, end, modalities=None,
                          enrolment_date=None):
    """ Returns a data completion array and a list of modality names.
    Parameters
    __________
    ptc: radar.Participant
    start: pd.Timestamp / np.datetime64
    end: pd.Timestamp / np.datetime64
    modalities: list of str
    enrolment_date: pd.Timestamp / np.datetime64

    Returns
    _______
    completion: np.ndarray
        First index corresponds to the modality given in the modalities
        list, the second index corresponds to hourly intervals between
        start and end.
    modalities: list of str
        A list of modalities, the same as input if given.
    """
    if modalities is None:
        modalities = sorted(list(ptc.data.keys()))
    completion = np.zeros((len(modalities), 1+(end-start)//HOUR), dtype=int)
    for i, dname in enumerate(modalities):
        if dname not in ptc.data:
            print(f'Participant {ptc.name} has no recorded {dname}')
            continue
        print(dname)
        completion[i, :] = data_completion_hourly(
                ptc.data[dname], start, end,
                enrolment_date=enrolment_date)
    return completion, modalities


def ptc_completion_plot(completion, modalities, start,
        ax=None, enrolment_date=None, cmap=CMAP):
    """ Plots the completion array as a heatmap
    Parameters
    __________
    completion: np.ndarray
        Completion 2d array for a participant
    modalities: list of str
        A list of the modalities, used to label the Y-axis
    start: pd.Timedelta / np.datetime64
    ax: matplotlib axes
    enrolment_date: pd.Timedelta / np.datetime64
    cmap: Colormap

    Returns
    _______
    ax: matplotlib axes
    """
    ax = sns.heatmap(completion, cmap=cmap, cbar=False,
                     xticklabels=672, ax=None)
    # Gaps
    N_x = completion.shape[1]
    N_y = completion.shape[0]
    for i in range(N_y-1):
        ax.hlines(i+1, 0, N_x, color='white', linewidth=2)

    # Outline
    ax.vlines(0, 0, N_y, color='black', linewidth=2)
    ax.vlines(N_x, 0, N_y, color='black', linewidth=2)
    ax.hlines(0, 0, N_x, color='black', linewidth=2)
    ax.hlines(N_y, 0, N_x, color='black', linewidth=2)

    # Enrolment
    if enrolment_date:
        enrolment_idx = (enrolment_date - start)//HOUR
        ax.vline(enrolment_idx-1, 0, N_y,
                 color='red', linewidth=0.5)

    # Labels
    ax.set_ylabel('Data topic')
    ax.set_xlabel('Date')
    xticks = ax.get_xticks()
    ax.set_xticklabels([(start + (tm * HOUR)).strftime('%y-%m-%d')
                        for tm in xticks], rotation=0)
    ax.set_yticklabels(modalities, rotation=0)
    return ax

if __name__ == '__main__':
    import os
    from datetime import date

    import matplotlib.pyplot as plt
    import radar
    from radar.util.projects.epi.parsers import RedcapCsv

    rc = RedcapCsv(csv_path='path/to/rc.csv', timezone='Europe/London')
    proj = radar.Project(name='EPI', path='path/to/project',
            info=rc.info(), datakw={'subdirs':['RADAR']})
    # the datakw 'subdirs' is used to find data folders inside subfolders
    # within a participant folder. E.g. RADAR in KCL EPI.
    modalities = [
            'android_empatica_e4_blood_volume_pulse',
            'android_empatica_e4_acceleration',
            'android_empatica_e4_electrodermal_activity',
    ] # Etc
    os.makedirs('reports/figures/completion_{}'.format(date.today()),
                exist_ok=True)
    for ptc in proj.participants:
        start = ptc.info['recording_start']
        end = ptc.info['recording_end'] + 12*HOUR
        fig = plt.figure(figsize=(12,6))
        completion, _ = ptc_completion_hourly(ptc, start, end, modalities)
        ptc_completion_plot(completion, modalities, start)
        plt.title('{}'.format(ptc.name))
        plt.savefig('reports/figures/completion_{}/{}.png'\
                .format(date.today(), ptc.name),
                    dpi=288, bbox_inches='tight')
        plt.close()



#!/usr/bin/env python3
import numpy as np
import pandas as pd
from .filters import butterworth
from typing import List, Tuple

SEC = pd.Timedelta(1, 's')

def calculate_drift(timeReceived: pd.Series) -> pd.Series:
    """ Calculate timedelta between timeRecieved and time.
    Parameters
    __________
    timeReceived: pandas.Series
        A pandas timestamp series with a timestamp index.

    Returns
    ________
    drift: pandas.Series
        A pandas int64 series showing the difference (ns)
        between index and input series.
    """
    return (timeReceived - timeReceived.index).astype('int64')


def drift_filter(delta: pd.Series, freq: float = 32) -> pd.Series:
    """ Filters the time drift series
    Finds the minimum value in rolling 120s sections.
    Params:
        delta (pd.Series[int64]): Series with timestamp index
        freq (float): Unused
    Returns:
        pd.Series[float64]: filtered drift
    """

    filt = delta.rolling('120s').min()
    start = filt.index[0]
    filt.loc[start:start+120*SEC] = filt.loc[start:start+120*SEC].min()
    return filt


def get_segments(series: pd.Series) -> List[Tuple[int, int]]:
    """ Get contiguous segments. Splits on gaps in the time index > 0.5sec
    Params:
        series (pd.Series)
    Returns:
        list: List of tuples (Segment indices)
    """
    def split_gaps(series):
        arr = series.index.astype(int)
        deriv = np.append(0, np.diff(arr))
        return np.where(np.abs(deriv) > 5e8)[0].tolist()

    gaps = split_gaps(series)
    gaps.append(len(series))
    segments = [(0, gaps[0])]
    for i in range(0, len(gaps)-1):
        segments.append((gaps[i], gaps[i+1]))
    return segments


def filter_segments(series, segments):
    out = series.copy()
    for i, ind in enumerate(segments):
        seg = series.iloc[slice(*ind)]
        out.iloc[slice(*ind)] = drift_filter(seg)
    return out


def lm_segments(series, segments):
    """ Fit linear models to each segment
    Currently unused
    """
    def fit_lm(series):
        x = series.index.values.astype('int64')
        y = series.values
        fit = np.polyfit(x, y, 1)
        return np.poly1d(fit)

    # zero = np.max([series.iloc[seg[0]] for seg in segments])
    # zero = np.max(zero, 0)
    fits = {'time': [0] * len(segments),
            'fit': [0] * len(segments)}
    for i, ind in enumerate(segments):
        seg = series.iloc[slice(*ind)]
        fits['time'][i] = seg.index[0]
        fits['fit'][i] = fit_lm(seg)
    return pd.DataFrame(data=fits).set_index('time')


def correct_drift(df, inplace=True):
    if not inplace:
        df = df.copy()
    drift = calculate_drift(df['timeReceived'])
    segments = get_segments(drift)
    filtered = filter_segments(drift, segments)
    df.index = df.index + filtered.astype('timedelta64[ns]')
    df = df.sort_index()
    return df



#!/usr/bin/env python3
import matplotlib.pyplot as plt

def group_plot(dataframe, ycols, groupby, xcol=None, **plotkwargs):
    groups = dataframe.groupby(groupby)[ycols]
    f = plt.figure()
    z = 1
    for label, group in groups:
        ax = plt.subplot(len(groups), 1, z)
        group.plot(y=ycols, x=xcol, ax=ax, **plotkwargs)
        z = z + 1
    plt.subplots_adjust(hspace=0)
    return f


def time_view(dataframes, timeframe, ycols, y2cols=None, xcol=None,
              titles=None, fig=None, **plotkwargs):
    """Plots data columns from Pandas dataframes at a certain time interval

    Parameters
    __________
    dataframes : list
        List of pandas dataframes with a timestamp index. (Alternatively, a
        shared x-axis column name supplied by xcol)
    timeframe: tuple
        Two strings containing the start and end date/time of the plot
    ycols: list
        List of list of the dataframe column names. Each element of the list
        corresponds to one subplot. Each element contains a list of  the names of
        columns for the corresponding dataframe.
    y2cols: list
        The same as ycols, but plotted on a second y-axis
    xcol: str
        An optional string specifying the x-axis column of each dataframe if
        the index is not used.
    titles: list
        List of strings for each subplot (dataframes) title.
    """
    z_max = len(dataframes)
    if not fig:
        fig = plt.figure()
    if not y2cols:
        y2cols = [False for i in range(z_max)]
    if not titles:
        titles = [False for i in range(z_max)]

    for z in range(z_max):
        ax = plt.subplot(z_max, 1, z+1)
        dataframes[z].loc[timeframe[0]:timeframe[1]].plot(y=ycols[z], secondary_y=y2cols[z], x=xcol, ax=ax,
                           **plotkwargs)
        if titles[z]:
            ax.set_title(titles[z])
        if z < z_max-1:
            ax.set_xlabel('')
            ax.set_xticks([])
            ax.set_xticklabels([])
    return fig



""" Commonly used functions the log -
This should probably be merged with generic.py, and the
log moved to it's own file
"""
import logging
from datetime import datetime
from . import config


def add_cli_log(logger: logging.Logger,
                formatter: logging.Formatter,
                level: int = logging.ERROR) -> None:
    """Create a CLI log of RADAR errors and above, used by default."""
    clilog = logging.StreamHandler()
    clilog.setLevel(level)
    clilog.setFormatter(formatter)
    logger.addHandler(clilog)


def add_file_log(logger: logging.Logger,
                 formatter: logging.Formatter,
                 level: int = logging.WARNING) -> None:
    """Create a file-based logger for warnings, debug, etc.
    Requires log_to_file config flag == True """
    filename = 'radar_{:%Y%m%d-%H%M}.log'.format(datetime.now())
    filelog = logging.FileHandler(filename)
    filelog.setLevel(level)
    filelog.setFormatter(formatter)
    logger.addHandler(filelog)


log = logging.getLogger('radarlog')
formatter = logging.Formatter(config.log.format)
add_cli_log(log, formatter)
if config.log.to_file:
    add_file_log(log, formatter)



#!/usr/bin/env python3
""" Generic IO functions
"""
from typing import List, Union
import os
import numpy as np
import pandas as pd
from ..generic import methdispatch


class FakeDatetimeArray():
    """ A fake numpy datetime array
    """
    ndim = 1
    _index_error = (
        'Only integer, slices (`:`), ellipsis (`...`), '
        ' integer arrays or boolean arrays are valid indices')

    def __init__(self, start: np.datetime64, length: int,
                 freq: Union[int, float, None] = None,
                 step: Union[int, float, None] = None):
        """
        step : int
            step size in seconds (1/freq)

        negative slices don't work
        """
        if step is None and freq is None:
            raise ValueError('Must provide either freq or step')
        if step:
            self.step = int(1e9) * step
            self.freq = 1/step
        else:
            self.freq = freq
            self.step = int(1e9 / self.freq)
        self.start = pd.Timestamp(start, 'ns').asm8
        self.length = length
        self.shape = (length,)
        self.dtype = np.dtype('datetime64[ns]')

    def __len__(self):
        return self.length

    @methdispatch
    def __getitem__(self, x):
        if x == Ellipsis:
            return self.arange(0, self.length, 1)
        raise IndexError(self._index_error)

    @__getitem__.register(slice)
    def __getslice__(self, x):
        start, stop, step = self.positive(x)
        return self.arange(start, stop, step)

    @__getitem__.register(tuple)
    def __gettuple__(self, x):
        return self.__getitem__(x[0])

    @__getitem__.register(int)
    def __getint__(self, x):
        x = self.positive(x)
        if 0 <= x < self.length:
            return self(x)
        raise IndexError(str(x) + ' is outside of 0-' + str(self.length))

    @__getitem__.register(list)
    def __getlist__(self, x):
        arr = None
        if isinstance(x[0], bool) and self.length == len(x):
            arr = np.where(x)
        elif isinstance(x[0], int):
            arr = np.array(x)
        if arr is not None:
            return self.from_arr(arr)
        raise IndexError(self._index_error)

    @__getitem__.register(np.ndarray)
    def __getnp__(self, x):
        if x.dtype == 'bool' and self.length == len(x):
            arr = np.where(x)
        else:
            arr = x
        if arr is not None:
            return self.from_arr(arr)
        raise IndexError(self._index_error)

    def from_arr(self, arr: np.ndarray) -> np.ndarray:
        """ Return datetime array from index array
        Params:
            arr (np.ndarray): Index array
        Returns:
            np.ndarray[np.datetime64]
        """
        return pd.DatetimeIndex(self.start + (arr * self.step))

    def __call__(self, x: int) -> np.ndarray:
        return self.start + (self.positive(x) * self.step)

    def arange(self, start: int, stop: int, step: int = 1) -> np.ndarray:
        """ Return evenly spaced numpy datetime array from
        start / stop / step index
        Params:
            start (int): Starting index of array
            stop (int): Stop index of array
            step (int): Step between index elements
        Returns:
            np.ndarray[np.datetime64]: Datetime array
        """
        dt_start = self(start)
        dt_stop = self(stop)
        dt_step = self.step * step
        return np.arange(dt_start, dt_stop, dt_step)

    @methdispatch
    def positive(self, x):
        """ Returns array length - x if x is less than 0
        Params:
            x (int): Array integer
        Returns:
            int: positive array integer
        """
        if x < 0:
            return self.length + x
        return x

    @positive.register(np.ndarray)
    def _(self, x):
        x[x < 0] = self.length + x[x < 0]
        if (x < 0).any():
            raise IndexError('Some indx outside of range')
        return x

    @positive.register(slice)
    def _(self, x):
        start = self.positive(x.start if x.start else 0)
        stop = self.positive(x.stop if x.stop else self.length)
        step = x.step if x.step else 1
        if start < 0:
            start = 0
        if stop < 0:
            stop = 0
        return start, stop, step



#!/usr/bin/env python3
import pandas as pd
import numpy as np

DAY = pd.Timedelta('1D')

class RedcapEpi():
    """ A class to parse information from an EPI Redcap CSV extract.
        The data dictionary can be found on the radar-base github.
    """
    def __init__(self, csv_path, timezone):
        """
        Parameters
        __________
        csv_path : str
            path of the csv file
        timezone : str / tz representation
            The timezone that the CSV has times recorded in.
        """
        self.df = pd.read_csv(csv_path)
        self.tz = timezone

    def seizures(self):
        """
        Returns
        _______
        labels : dict
            A dict where each key is a patient code with another dict as
            a value. That dict has keys representing labels ('Seizure') and
            values of a dataframe with start and end times.
        """
        df = self.df.copy()
        cols = ('clin_start', 'clin_end', 'eeg_start', 'eeg_end')
        for c in cols:
            df[c] = pd.to_datetime(df.rep_seizure_date + 'T' +
                    df['rep_seizure_' + c].str.replace('.', ':'))
        df['start'] = df[['clin_start', 'eeg_start']].min(axis=1)
        df['end'] = df[['clin_end', 'eeg_end']].min(axis=1)
        df.start = df.start.dt.tz_localize(self.tz)
        df.end = df.end.dt.tz_localize(self.tz)
        labels = {}
        for i in range(1, max(self.df.record_id) + 1):
            ptc = df[df.record_id == i]
            code = ptc.patient_code.iloc[0]
            seizures = ptc.iloc[1:]
            labels[code] = {}
            labels[code]['Seizure'] = \
                    seizures[['start', 'end']].reset_index(drop=True)
        return labels

    def info(self):
        """
        Returns
        _______
        info : dict
            A dict where each key is a patient code with another dict as
            a value. That dict has keys corresponding to patient information
            fields - enrolment date, recording start, and recording end
        """
        info = {}
        for i in range(1, max(self.df.record_id) + 1):
            ptc = self.df[self.df.record_id == i]
            code = ptc.patient_code.iloc[0]
            enrolled = ptc.enrolment_date.iloc[0]
            recording_start = ptc.start_date.iloc[0]
            recording_end = ptc.recording_end_date.iloc[0]
            info[code] = {}
            info[code]['enrolment_date'] = \
                    pd.to_datetime(enrolled).tz_localize(self.tz)
            info[code]['recording_start'] = \
                    pd.to_datetime(recording_start).tz_localize(self.tz)
            info[code]['recording_end'] = \
                    pd.to_datetime(recording_end).tz_localize(self.tz) + DAY
        return info




#!/usr/bin/env python3
from typing import Callable, Dict
import dateutil
from datetime import datetime
import numpy as np
import pandas as pd


DT_MULT = int(1e9)


class RadarCsvReader():
    def __init__(self, dtypes=None, column_processors=None,
                 has_array_fields=True, timecols=tuple(),
                 replace_dots=True):
        self.dtypes = dtypes
        self.column_processors = column_processors if column_processors else {}
        self.has_array_fields = has_array_fields
        self.timecols = timecols
        self.replace_dots = replace_dots

    def __call__(self, fn, fs, compression='gzip'):
        f = fs.open(fn)
        df = pd.read_csv(f, dtype=self.dtypes, parse_dates=False, compression=compression)
        f.close()
        df.columns = ['.'.join(c.split('.')[1:]) for c in df.columns]
        for c, processor in self.column_processors.items():
            df[c] = processor(df[c])
        if self.has_array_fields:
            df = _melt_array_fields(df)
        for c in self.timecols:
            df[c] = date_parser(df[c])
        index = choose_index(df.columns)
        if index:
            df = df.set_index(index)
        if 'userId' in df.columns:
            df['userId'] = df['userId'].apply(lambda x: x[:36])
        if self.replace_dots:
            df.columns = [c.replace('.', '_') for c in df.columns]
        return df


def choose_index(columns):
    for c in ['time', 'dateTime', 'timeReceived']:
        if c in columns:
            return c
    return None


def find_array_columns(columns):
    def is_array_column(c):
        c_split = c.split('.')
        if len(c_split) > 2:
            if c_split[-2].isdigit():
                return True
        return False
    id_vars = []
    arr_vars = []
    for c in columns:
        if is_array_column(c):
            arr_vars.append(c)
        else:
            id_vars.append(c)
    return id_vars, arr_vars


def _melt_array_fields(df):
    def field_name(x):
        x.pop(-2)
        return '.'.join(x)

    def melt_row(row, ids):
        melt = row.melt(id_vars=ids)
        melt['arrid'] = [(lambda x: x[-2])(x)
                         for x in melt['variable'].str.split('.')]
        melt['field'] = [field_name(x)
                         for x in melt['variable'].str.split('.')]
        col_data = [(x[0], x[1]['value'].values)
                    for x in melt.groupby('field')]
        col_data.append(('arrid', melt[melt['field'] == melt['field'][0]]['arrid'].values))
        melt = pd.DataFrame([x[1][ids].iloc[0] for x in melt.groupby('arrid')])
        melt = melt.reset_index(drop=True)
        for col, data in col_data:
            melt[col] = data
        return melt
    id_vars, arr_vars = find_array_columns(df.columns)
    if any(id_vars) and any(arr_vars):
        df = pd.concat([melt_row(df[i:i+1], id_vars)
                        for i in range(len(df))], sort=True)
    return df


def date_parser(series):
    """Convert an array of floats or strings to a datetime object.
    Params:
        series (pd.Series[float/str])
    Returns:
        np.ndarray[M8[ns]] / datetime.datetime
    """
    series = series.values
    if type(series[0]) == float:
        series = series.astype(float)
    if series.dtype == 'float':
        return (series * 1e9).astype('M8[ns]')
    return pd.to_datetime(series)


def is_numeric(series):
    """ Checks whether pandas series dtype is a float or integer.
    Params:
        series (pd.Series): Pandas series to check
    Returns:
        bool
    """
    return series.dtype == 'float' or series.dtype == 'int'


def convert_to_datetime(series):
    if is_numeric(series):
        return pd.DatetimeIndex(DT_MULT * series, tz='UTC')
    return pd.DatetimeIndex(pd.to_datetime(series))



import re
import collections
from functools import lru_cache, singledispatch, update_wrapper


class Registry(dict):
    def register(self, name, val=None):
        def wrapper(func):
            self[name] = func
        if val is None:
            return wrapper
        else:
            self[name] = val


def update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def methdispatch(func):
    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    dispatcher = singledispatch(func)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


@lru_cache(8)
def re_compile(pattern):
    return re.compile(pattern)



""" IO operations for SQL databases - currently Postgres specific
Currently unsafe / doesn't sanatize inputs. Should work in psycopg2 2.8
"""
import time
from typing import Callable, Union, Any, List, Tuple
from io import StringIO
from functools import wraps, partial
from itertools import chain

import psycopg2
from psycopg2.extensions import connection as Connection
from psycopg2.pool import AbstractConnectionPool

import pandas as pd
from ..log import log


class PostgresIOPool():
    """ Class version of module
    """
    def __init__(self, pool: AbstractConnectionPool):
        self.pool = pool

    def df_to_psql(self, table: str, df: pd.DataFrame):
        """ Insert a pandas DataFrame into a Postgres table using a
        StringIO intermediary CSV to COPY FRO
        Args:
            table (str): postgres table name
            df (pandas.DataFrame): dataframe to insert
        Returns:
            None
        """
        conn = self.pool.getconn()
        return df_to_psql(conn, table, df)

    def create_table_like(self, new_table: str, existing_table: str):
        """ Create an SQL table based on another existing table
        Args:
            new_table (str): The name of the table being created
            existing_table (str): The name of the table to copy
        Returns:
            None
        """
        conn = self.pool.getconn()
        return create_table_like(conn, new_table, existing_table)

    def drop_table(self, table: str):
        """ Drop a table (dynamic name)
        Args:
            table (str): table name
        """
        conn = self.pool.getconn()
        drop_table(conn, table)

    def upsert_from_table(self, source_table: str,
                          target_table: str, columns: List[str],
                          constraints: Tuple[str] = ('userid', 'time')) -> None:
        """ Copy between tables and upsert all columns.
        Currently potentially unsafe - doesn't sanitize table or
        column names - should be fixed once psycopg2 2.8 comes out
        Args:
            source_table (str): Name of table to copy from
            target_table (str): Name of table to copy into
            columns (list): List of column names
            constraints (list): List of UNIQUE CONSTRAINT
                columns in sql table
        """
        conn = self.pool.getconn()
        return upsert_from_table(conn, source_table,
                                 target_table, columns, constraints)

    def df_upsert_psql(self, table: str, df: pd.DataFrame) -> None:
        """ Upsert a Dataframe into a Postgres table through:
        DataFrame -> StringIO -> Temp table -> table
        Args:
            table (str): Table to upsert data into
            df (pd.DataFrame): Dataframe containing data to upsert from
        """
        conn = self.pool.getconn()
        df_upsert_psql(conn, table, df)


def try_sql(func: Callable[[Connection, Any], Any]) -> Callable:
    """ Wrap a function that executes an SQL query in try-except.
    Cancels the connection on error, commits on success.
    """
    @wraps(func)
    def wrapper(conn: Connection, *args, **kwargs) -> Union[Any, str]:
        """ Try-except wrapper. Returns func output on success"""
        try:
            out = func(conn, *args, **kwargs)
        except psycopg2.Error as err:
            log.error('SQL error with func %s, args %s', func, args)
            curs = conn.cursor()
            curs.execute("ROLLBACK")
            conn.commit()
            raise err
        conn.commit()
        return out
    return wrapper


@try_sql
def df_to_psql(conn: Connection, table: str, df: pd.DataFrame) -> None:
    """ Insert a pandas DataFrame into a Postgres table using a
    StringIO intermediary CSV to COPY FROM
    Args:
        conn: psycopg2 connection
        table (str): postgres table name
        df (pandas.DataFrame): dataframe to insert
    Returns:
        None
    """
    cur = conn.cursor()
    f = StringIO()
    df.to_csv(f, na_rep='\\N', sep=';')
    f.seek(0)
    cols = f.readline().strip().split(';')
    cur.copy_from(f, table, sep=';', columns=cols)
    cur.close()


@try_sql
def create_table_like(conn: Connection,
                      new_table: str,
                      existing_table: str):
    """ Create an SQL table based on another existing table
    Args:
        conn: psycopg2 connection to the postgres database
        new_table (str): The name of the table being created
        existing_table (str): The name of the table to copy
    Returns:
        None
    """
    cur = conn.cursor()
    cmd = (
        'CREATE TABLE {} (like {} '
        'INCLUDING DEFAULTS '
        'INCLUDING CONSTRAINTS '
        'INCLUDING INDEXES)').format(new_table, existing_table)
    cur.execute(cmd)
    cur.close()
    conn.commit()


@try_sql
def drop_table(conn: Connection, table: str):
    """ Drop a table (dynamic name)
    Args:
        conn: psycopg2 connection
        table (str): table name
    """
    cur = conn.cursor()
    cmd = 'DROP TABLE {}'.format(table)
    cur.execute(cmd)
    cur.close()


@try_sql
def upsert_from_table(conn, source_table: str, target_table: str,
                      columns: List[str],
                      constraints: List[str] = ['userid', 'time']) -> None:
    """ Copy between tables and upsert all columns.
    Currently potentially unsafe - doesn't sanitize table or
    column names - should be fixed once psycopg2 2.8 comes out
    Args:
        conn: psycopg2 connection
        source_table (str): Name of table to copy from
        target_table (str): Name of table to copy into
        columns (list): List of column names
        constraints (list): List of UNIQUE CONSTRAINT
            columns in sql table
    """
    columns = [c.lower() for c in columns if c not in ('userid', 'time')]
    cur = conn.cursor()
    cmd_string = (
        'INSERT INTO {}'
        ' SELECT * FROM {}'
        ' ON CONFLICT ( ' + ', '.join(['{}' for c in constraints]) + ')'
        ' DO UPDATE SET '
    )
    cmd_string += ', '.join(['{} = {}' for c in columns])
    cmd = cmd_string.format(
        target_table, source_table,
        *[c for c in constraints],
        *list(chain.from_iterable((
            (c, 'EXCLUDED.' + c)) for c in columns)))
    cur.execute(cmd)


def df_upsert_psql(conn: Connection, table: str, df: pd.DataFrame,
                   constraints: List[str] = ['userid', 'time']) -> None:
    """ Upsert a Dataframe into a Postgres table through:
    DataFrame -> StringIO -> Temp table -> table
    Args:
        conn: psycopg2 connection
        table (str): Table to upsert data into
        df (pd.DataFrame): Dataframe containing data to upsert from
    """
    idx_name = df.index.name
    if idx_name:
        df = df.reset_index()
    df = df.drop_duplicates(subset=constraints)
    if idx_name:
        df = df.set_index(idx_name)
    columns = [df.index.name] + list(df.columns)
    curr_time = str(time.time()).replace('.', '_')
    tmp_table = table + '_staging_' + curr_time
    create_table_like(conn, tmp_table, table)
    try:
        df_to_psql(conn, tmp_table, df)
        upsert_from_table(conn, tmp_table, table, columns, constraints)
    finally:
        drop_table(conn, tmp_table)


def df_upsert_with_pool(pool, table, df, constraints):
    conn = pool.getconn()
    try:
        df_upsert_psql(conn, table, df, constraints)
    finally:
        conn.close()
        pool.putconn(conn)


def dask_upsert_psql(pool, table, ddf, constraints=('userid', 'time')):
    """ Upsert a dask dataframe into SQL table
    Args:
        pool: psycopg2 connection pool
        table (str): table name
        ddf (dask dataframe)
    Returns:
        Dask delayed object
    """
    upsert_with_pool = partial(df_upsert_with_pool, pool=pool,
                               constraints=constraints, table=table)
    return ddf.map_partitions(upsert_with_pool, meta=(None, None))



#!/usr/bin/env python3
import glob
import yaml
import requests
from collections import OrderedDict
from packaging import version
from ..defaults import config


class ProjectSpecs(dict):
    """
    A dictionary class to hold all YML specifications relating to a
    RADAR project
    Parameters
    __________
    device_specifications : list of DeviceSpec
        A list of device specification objects
    """
    def __init__(self, device_specifications):
        self.devices = {dev.name: dev for dev in device_specifications}
        for dev in self.devices.values():
            self.update(dev)

    def __repr__(self):
        repr_string = 'Specification dictionary with:\n' + \
                      'Devices: {}\n'.format(', '.join(self.devices)) + \
                      'Specifications: {}'.format(', '.join(self))
        return repr_string


class DeviceSpec(dict):
    """
    A dictionary class to store RADAR YML specifications
    Parameters
    _________
    yml: dict
        A dictionary loaded with yaml.load(file)
    """
    def __init__(self, yml: dict):

        for attr, val in yml.items():
            if not isinstance(val, str):
                continue
            setattr(self, attr, val)

        for attr in ('vendor', 'model', 'version'):
            if not hasattr(self, attr):
                setattr(self, attr, '')

        self.name = '_'.join((self.vendor, self.model))

        super(DeviceSpec, self).__init__([(mdl['topic'], ModalitySpec(mdl))
              for mdl in yml['data'] if ('data' in yml and 'topic' in mdl)])


class ModalitySpec(dict):
    """
    A dictionary class to store the modalities of a RADAR specification.
    """

    def __init__(self, modal:dict):
        for attr, val in modal.items():
            if attr is not 'fields':
                setattr(self, attr, val)
        if 'fields' in modal:
            self.update([('value.' + field['name'], FieldSpec(field))
                         for field in modal['fields']])

    def __repr__(self):
        repr_string = '{} modality of type "{}" '.format(self.topic,
                       self.type if hasattr(self, 'type') else 'UNKNOWN') +\
                      'with fields: ' + ', '.join(self)
        return repr_string

    def group_fields(self, ):
        return -1

    def _type_columns(self, coltype):
        return [name for name, col in self.items()
                if 'type' in col and
                col['type'] == coltype]

    def timecols(self):
        return self._type_columns('TIMESTAMP')

    def timedeltas(self):
        return self._type_columns('DURATION')


class FieldSpec(OrderedDict):
    """
    A class to store fields of modalities in RADAR specifications.
    """
    def __repr__(self):
        repr_string = '"{}" field: (({}))'.format(self['name'],
                '), ('.join([': '.join((k, v)) for k, v in self.items()]))
        return repr_string


def specifications_from_directory(path=config.specifications.dir):
    files = glob.glob(path + '/**/*.yml', recursive=True)
    device_specs = {}
    for fn in files:
        with open(fn) as f:
            spec = DeviceSpec(yaml.load(f, yaml.SafeLoader))
        if (spec.name not in device_specs or
                (version.parse(spec.version) >
                 version.parse(device_specs[spec.name].version))):
            device_specs[spec.name] = spec
    return ProjectSpecs(device_specs.values())


def specifications_from_github(repo_owner, repo_name, sha):
    url = 'https://api.github.com/repos/{}/{}/contents/specifications/'\
        .format(repo_owner, repo_name)

    rawurl = 'https://raw.githubusercontent.com/{}/{}/{}/'\
        .format(repo_owner, repo_name, sha)

    def ls(relpath):
        requrl = url + relpath
        return requests.get(requrl,
                            params={'ref': 'master' if sha is None else sha})

    def dl(relpath):
        return requests.get(rawurl + relpath).content

    folders = ('active', 'connector', 'monitor', 'passive')
    files = [f['path'] for folder in folders for f in ls(folder).json()]
    ymls = [yaml.load(dl(spec), yaml.SafeLoader) for spec in files]
    return ProjectSpecs([DeviceSpec(y) for y in ymls])



""" Cleaning and aggregation functions for RADAR data
"""
from functools import wraps
import numpy as np
import pandas as pd
from . import validators
from ..generic import Registry


CLEANERS = Registry()


def last(x):
    """ Returns the last element of a series.
    Params:
        x (pd.Series)
    Returns:
        scalar of x.dtype
    """
    return x.iloc[-1]


def validate(validator):
    """ Wrapper to run a validator on dataframe input."""
    def decorator(func):
        @wraps(func)
        def wrap(df, *args, **kwargs):
            df = df[validator(df)]
            return func(df, *args, **kwargs)
        return wrap
    return decorator


@CLEANERS.register('android_phone_contacts')
@validate(validators.android_phone_contacts)
def android_phone_contacts(df: pd.DataFrame) -> pd.DataFrame:
    """ Returns daily aggregate of valid phone_contacts.
    Sampling frequency of phone_contacts can be high -
    but only really need daily. There are also occasional bugs
    (see validator.android_phone_contacts)
    Params:
        df (pd.DataFrame): phone_contacts dataframe
    Returns:
        pd.DataFrame: daily aggregate of phone_contacts
    """
    aggregations = {
        'contactsAdded': sum,
        'contactsRemoved': sum,
        'contacts': lambda x: np.median(x).astype(df['contacts'].dtype)
    }

    for col in df.columns:
        if col not in aggregations:
            aggregations[col] = last

    df = df[validators.android_phone_contacts(df)]
    return df.groupby(df.index.floor('1D')).agg(aggregations)


@CLEANERS.register('questionnaire_phq')
@validate(validators.questionnaire_phq8)
def questionnaire_phq8(df: pd.DataFrame) -> pd.DataFrame:
    """ Returns a dataframe with only validated answers
    Params:
        df (pd.DataFrame): melted PHQ8 dataframe
    Returns:
        pd.DataFrame: valid PHQ8 answers
    """
    return df


@CLEANERS.register('questionnaire_rses')
@validate(validators.questionnaire_rses)
def questionnaire_rses(df: pd.DataFrame) -> pd.DataFrame:
    """ Returns a dataframe with only validated answers
    Params:
        df (pd.DataFrame): melted RSES dataframe
    Returns:
        pd.DataFrame: valid RSES answers
    """
    return df


@CLEANERS.register('connect_fitbit_sleep_stages')
def connect_fitbit_sleep_stages(df: pd.DataFrame) -> pd.Series:
    """ Cleans fitbit_sleep_stages dataframe
    Maps UNKNOWN -> AWAKE (bug in early connector)
    Params:
        df (pd.DataFrame): sleep stages dataframe
    Returns:
        pd.DataFrame
    """
    df.loc[df['level'] == 'UNKNOWN', 'level'] = 'AWAKE'
    return df



#!/usr/bin/env python3
from functools import lru_cache
import pandas as pd
import numpy as np
from ....generic import update


INFO = (('study',
            ['study',
              'enrolment_date',
              'visit',
              'month']),
        ('study_devices',
            ['fitbit_serial',
             'phone_status',
             'phone_brand',
             'phone_model',
             'old_mobile_device',
             'phone_imei_number',
             'android_version',
             'contract_provided',
             'contract_start',
             'armt_installed',
             'armt_version',
             'prmt_installed',
             'thincit_installed',
             'mindstrong_installed']),
        ('socio_demographics',
            ['dob',
             'age',
             'gender',
             'gender_at_birth',
             'marital_status',
             'living_together',
             'children',
             'no_of_children',
             'children_live_with_you',
             'youngest_child_age',
             'country',
             'eduction_finish_age', # [sic]
             'drop_school_1',
             'qualifications',
             'degree_1',
             'benefits',
             'income_gbp',
             'income_contribution',
             'accommodation',
             'accommodation_satisfaction']))

class RedcapMDD():
    """
    A class to parse information from an EPI Redcap CSV extract.
    The data dictionary can be found on the radar-base github.
    """
    def __init__(self, csv_path, timezone=None):
        """
        Params:
            csv_path (str): path of the RedCap csv extract file
            timezone (str): The timezone that was used to record datetimes
        """
        df = pd.read_csv(csv_path, sep=None, engine='python')
        self._id_cols = ['subject_id', 'record_id']
        for c in ['subject_id', 'human_readable_id']:
            df[c].fillna(method='ffill', inplace=True)
        for c in df.columns:
            if 'timestamp' in c:
                df[c] = (pd.to_datetime(df[c], errors='coerce')
                         .dt.tz_localize(timezone))
        df['enrolment_date'] = pd.to_datetime(df['enrolment_date'])
        df['enrolment_date'] = df['enrolment_date'].dt.tz_localize(timezone)
        self.df = df
        self.tz = timezone

    def _subset(self, columns):
        id_cols = ['subject_id', 'record_id']
        out = self.df[columns].copy()
        out = out.dropna(how='all')
        for c in id_cols:
            out.insert(0, c, self.df[c].copy())
        return out

    @property
    @lru_cache(1)
    def ids_sr(self):
        cols = [c for c in self.df.columns if
                c[:4] == 'ids_' and c[4].isnumeric()]
        cols.extend([c for c in self.df.columns if
                     'ids_sr' in c])
        return self._subset(cols)

    @property
    @lru_cache(1)
    def info(self):
        info = {}
        df = self.df.drop_duplicates('record_id', keep='first')
        for sid, r in df.groupby('subject_id'):
            info[sid] = {}
            info[sid]['record_id'] = r.iloc[0]['record_id']
            for cat, fields in INFO:
                fields = r.columns.intersection(fields)
                info[sid][cat] = r.iloc[0][fields].to_dict()
        return info


def _ids_sr_severity(score):
    """Returns a depression severity from a IDS_SR score.
    Scale taken from: http://www.ids-qids.org/interpretation.html
    Parameters
    __________
    score : numeric
        IDS-SR total score
    Returns
    ______
    severity : float
        Severity according to above scale. Float to allow NaN if score is NaN.
    """
    if score >= 49:
        severity = 4.0
    elif score >= 39:
        severity = 3.0
    elif score >= 26:
        severity = 2.0
    elif score >= 14:
        severity = 1.0
    elif np.isnan(score):
        severity = score
    else:
        severity = 0.0
    return severity



#!/usr/bin/env python3
from typing import Dict
import os
import json
import glob
import numpy as np
import pandas as pd
import avro.schema
import fsspec
from ..defaults import config, schemas
import collections
from functools import lru_cache


AVRO_NP_TYPES = {
    'null': 'object',
    'boolean': 'bool',
    'int': 'Int32',
    'long': 'Int64',
    'float': 'float32',
    'double': 'float64',
    'bytes': 'bytes',
    'string': 'object',
    'enum': 'object',
}
DATETIME_FIELDS = {'time', 'timeReceived', 'dateTime', 'timeCompleted',
                    'timeNotification',
                   'answers.\d+.startTime', 'answers.\d+.endTime'}
TIMEDELTA_FIELDS = {'duration'}
_NAMES = avro.schema.Names()


class RadarSchema():
    def _get_timedelta_columns(self):
        def get_dtype(doc):
            dtype = 'timedelta64[s]'
            if 'milli' in doc:
                dtype = 'timedelta64[ms]'
            elif 'micro' in doc:
                dtype = 'timedelta64[us]'
            return dtype
        return {name: get_dtype(field.avro.doc.lower())
                for name, field in self.fields.items()
                if 'duration' in name.lower()
                or 'interval' in name.lower()}

    def __init__(self, fullname, name, namespace, fields):
        self.fullname = fullname
        self.name = name
        self.namespace = namespace
        self.fields = fields
        field_names = set(self.fields)
        self.timecols = [drop_arrid(name) for name in
                         field_names.intersection(DATETIME_FIELDS)]
        self.timedeltas = self._get_timedelta_columns()
        self.dtypes = {'value.' + k: v.nptype for k, v in self.fields.items()
                       if v.nptype is not None}

    def __eq__(self, other):
        if isinstance(other, RadarSchema):
            return (self.fields == other.fields and
                    self.fullname == other.fullname)
        return False

    def from_avro(avro_schema: avro.schema.RecordSchema):
        fullname = avro_schema.fullname
        name = avro_schema.name
        namespace = avro_schema.namespace
        fields = parse_schema_fields(avro_schema)
        return RadarSchema(fullname=fullname, name=name,
                           namespace=namespace, fields=fields)


def drop_arrid(name):
    spl = [s for s in name.split('.') if s != '\d+']
    return '.'.join(spl)

class SchemaField():
    def convert_type(self, dtype):
        if isinstance(dtype, avro.schema.UnionSchema):
            list_dtype = dtype.to_json()
            if 'null' in list_dtype:
                list_dtype.remove('null')
            if len(list_dtype) == 1:
                if list_dtype[0] in ('int', 'long'):
                    return 'Int64'
                elif list_dtype[0] == 'boolean':
                    return 'object'
                elif isinstance(dtype.schemas[1], avro.schema.ArraySchema):
                    return AVRO_NP_TYPES.get(dtype.schemas[1].items.type)
                elif isinstance(dtype.schemas[1], avro.schema.EnumSchema):
                    symbols = dtype.schemas[1].symbols
                    return pd.api.types.CategoricalDtype(symbols)
                else:
                    return AVRO_NP_TYPES.get(list_dtype[0])
            else:
                return None
        elif isinstance(dtype, avro.schema.EnumSchema):
            return pd.api.types.CategoricalDtype(dtype.symbols)
        elif isinstance(dtype, avro.schema.ArraySchema):
            return AVRO_NP_TYPES.get(dtype.items.type)
        return AVRO_NP_TYPES.get(dtype.type)
    def __init__(self, field: avro.schema.Field):
        self.avro = field
        self.nptype = self.convert_type(field.type)


def parse_schema_fields(schema: avro.schema.RecordSchema, namespace=''):
    def get_schema_if_record(field: avro.schema.Field):
        if field.type.type == 'record':
            return field.type
        if field.type.type == 'union':
            for r in field.type.schemas:
                if r.type == 'record':
                    return r
    def get_schema_if_array(field: avro.schema.Field):
        if field.type.type == 'array':
            return field.type
        if field.type.type == 'union':
            for r in field.type.schemas:
                if r.type == 'array':
                    return r
    out: Dict[str, SchemaField] = {}
    for field in schema.fields:
        rec = get_schema_if_record(field)
        arr = get_schema_if_array(field)
        name = namespace + field.name
        if rec:
            out.update(parse_schema_fields(rec, name + '.'))
        elif arr:
            if isinstance(arr.items, avro.schema.RecordSchema):
                out.update(parse_schema_fields(arr.items, name + '.\d+.'))
            else:
                out[name + '.\d+'] = SchemaField(field)
        else:
            out[name] = SchemaField(field)
    return out


def schemas_from_commons(path, names=_NAMES):
    out: Dict[str, RadarSchema] = {}
    excp_schemas = []
    spec = fsspec.utils.infer_storage_options(path)
    path = spec.pop('path')
    fs = fsspec.filesystem(**spec)
    schema_paths = fs.glob(path + fs.sep + '**.avsc')
    for path in schema_paths:
        try:
            scm = schema_from_file(path, names=names)
            if scm is not None:
                out[scm.fullname] = scm
        except avro.schema.SchemaParseException as exc:
            excp_schemas.append(path)
    for path in excp_schemas:
        scm = schema_from_file(path, names=names)
        if scm is not None:
            out[scm.fullname] = scm
    return out


def schema_from_file(f, names=_NAMES):
    with open(f) as of:
        data = json.load(of)
    try:
        avro_scm = avro.schema.SchemaFromJSONData(data, names=names)
    except avro.schema.SchemaParseException as exc:
        del names.names[data['namespace'] + '.' + data['name']]
        raise exc
    if isinstance(avro_scm, avro.schema.RecordSchema):
        return RadarSchema.from_avro(avro_scm)



import numpy as np
import pandas as pd
import dask.dataframe as dd
from ..common import log


def data_completion(ddf, start, end, freq='1h', enrolment_date=None):
    """ Returns a vector with elements representing whether data is
    present at resolution [freq].
    Parameters
    __________
    ddf: dask.dataframe
        A delayed dataframe as produced through radar.io.load().
        Must have a DatetimeIndex.
    start: pd.Timestamp / np.datetime64
    end: pd.Timestamp / np.datetime64
    freq: str
        Timedelta string determining the frequency (i.e. resolution) of the
        completion check. Anything other than the default '1h' resamples the
        data which may take significantly more time.
    enrolment_date: pd.Timestamp / np.datetime64 (optional)
        Hours before this date are assigned -1.

    Returns
    _______
    v: np.ndarray
        Array where 1 corresponds to present data, 0 to missing data,
        and -1 to prior to enrolment.
    """
    vec = pd.Series(index=pd.date_range(start=start, end=end, freq=freq),
                    dtype=int)
    if vec.empty or ddf is None:
        return []

    if freq == '1h':
        idx = pd.DatetimeIndex(ddf.divisions)
        idx = idx[np.logical_and(idx > vec.index[0], idx < vec.index[-1])]
        vec.loc[idx] = 1
    else:
        resampled = ddf.resample(freq).first().dropna()
        resampled = resampled.compute()
        for i, v in vec.iteritems():
            if i in resampled.index:
                vec.loc[i] = 1

    if enrolment_date:
        vec.loc[:enrolment_date] = -1
    return vec.values


def ptc_completion(ptc, start, end, freq='1h',
                   modalities=None, enrolment_date=None):
    """ Returns a data completion array and a list of modality names.
    Parameters
    __________
    ptc: radar.Participant
    start: pd.Timestamp / np.datetime64
    end: pd.Timestamp / np.datetime64
    freq: str
        Timedelta string determining the frequency (i.e. resolution) of the
        completion check. Anything other than the default '1h' resamples the
        data which may take significantly more time.
    modalities: list of str
    enrolment_date: pd.Timestamp / np.datetime64

    Returns
    _______
    completion: np.ndarray
        First index corresponds to the modality given in the modalities
        list, the second index corresponds to hourly intervals between
        start and end.
    modalities: list of str
        A list of modalities, the same as input if given.
    """
    if modalities is None:
        modalities = sorted(list(ptc.data.keys()))
    completion = np.zeros((len(modalities), 1+(end-start)//pd.Timedelta(freq)), dtype=int)
    for i, dname in enumerate(modalities):
        if dname not in ptc.data or not isinstance(ptc.data[dname], dd.DataFrame):  # or len(ptc.data[dname]) == 0:
            log.debug("Participant %s has no recorded %s", ptc.name, dname)
            continue
        log.debug("Processing %s for participant %s", dname, ptc.name)
        completion[i, :] = data_completion(ptc.data[dname], start, end, freq, enrolment_date=enrolment_date)
    return completion, modalities


def completion_idx_has_data(completion, idx, modalities=None, filter=None, requirement_function=any):
    """
    Parameters
    __________
    completion: np.ndarray
        Completion 2d array for a participant
    idx: int or slice
        Either the direct index of the data where to check, or an index slice of a subset of the data to check.
        If a slice is given, all the indices described by the slice need to have data in the completion.
    modalities: list of str
        The list of modalities associated with the completion array. Only used in combination with filter.
    filter: str
        A simple string filter for modalities.
        If both this and modalities are given, only those modalities are checked that include the filter as a substring.
    requirement_function: func
        The function that poses the requirement for a positive result, with respect to the modality-dimension.

    Returns
    _______
    b: boolean
        True if the requirement is fulfilled, False otherwise.
    """
    # TODO: Can this be any function that takes an iterable and returns a boolean? Can that be checked?
    if requirement_function != any and requirement_function != all:
        log.critical("completion_idx_has_data(): argument 'requirement_function' must be any() or all(). Default: any().")
        raise NotImplementedError("Currently only one of the built-ins any() and all() are allowed as requirement_function")

    idx_data = []
    if not isinstance(idx, slice): idx = slice(int(idx), int(idx)+1)
    if modalities is None or filter is None:
        idx_data = [ all(d[idx]) for d in completion ]
    else:
        mod_idx = [ i for i,v in enumerate(modalities) if filter in v ]
        idx_data = [ all(d[idx]) for i,d in enumerate(completion) if i in mod_idx ]

    return requirement_function(idx_data)



#!/usr/bin/env python3
from setuptools import setup, find_packages

setup(name='radarstudy',
      version='0.4.2',
      description='Offline processing and visualisation tools for use with data from RADAR studies',
      url='https://github.com/RADAR-base/pyRADAR-processing',
      author='Callum Stewart',
      author_email='callum.stewart@kcl.ac.uk',
      packages=find_packages(),
      install_requires=[
          'numpy',
          'pandas',
          'dask[complete]',
          'requests',
          'pyyaml>=5.1',
          'avro-python3'
          ],
      include_package_data=True,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: Apache Software License",
          "Intended Audience :: Science/Research"
          ]
)



#!/usr/bin/env python3
"""A module containing the base classes for RADAR.

Classes corresponding to the project, participant, and data
"""
import textwrap
from collections.abc import MutableMapping, KeysView, ItemsView, ValuesView
from datetime import datetime
from fsspec import filesystem
from fsspec.utils import infer_storage_options
import numpy as np
import pandas as pd
import dask.dataframe as dd
from dask import delayed
from .defaults import schemas, specifications
from .io.radar import RadarCsvReader


class RadarObject(MutableMapping):
    _keys_view = KeysView
    _items_view = ItemsView
    _values_view = ValuesView

    def __init__(self, path, populate=False, parent=None, name=None, **kwargs):
        self.parent = parent
        self._subobject_class = RadarObject
        fs = kwargs.get('fs')
        specs = infer_storage_options(path)
        self.path = specs.pop('path')
        if fs is None:
            fs = filesystem(**specs)
        self.fs = fs
        self.path = self.fs.info(self.path)['name'].rstrip(self.fs.sep)
        self.name = name if name else self.path.strip(self.fs.sep).split(self.fs.sep)[-1]
        self.store = dict()
        if populate:
            self._populate_store(populate=True)
        self._populated = populate

    def _list_subobjects(self):
        return self.fs.listdir(self.path)

    def _populate_store(self, populate=False):
        paths = self._list_subobjects()
        for p in paths:
            if p['type'] == 'directory':
                name = p['name'].rstrip(self.fs.sep)
                basename = name.split(self.fs.sep)[-1]
                self.store[basename] = \
                    self._subobject_class(name, parent=self, fs=self.fs)
        self._populated = True

    def __setitem__(self, key, val):
        self.store[key] = val

    def __getitem__(self, key):
        if not self._populated:
            self._populate_store()
        split_key = key.split('/')
        if len(split_key) == 1:
            return self.store[key]
        else:
            base_key = split_key[0]
            sub_key = '/'.join(split_key[1:])
            return self.store[base_key][sub_key]

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def keys(self):
        if not self._populated:
            self._populate_store()
        return self._keys_view(self)

    def items(self):
        if not self._populated:
            self._populate_store()
        return self._items_view(self)

    def values(self):
        if not self._populated:
            self._populate_store()
        return self._values_view(self)

    def _ipython_key_completions_(self):
        return list(self.keys())

    def _get_attr_or_parents(self, attr):
        if hasattr(self, attr) and getattr(self, attr) is not None:
            res = getattr(self, attr)
        elif getattr(self, 'parent') is None:
            res = None
        else:
            res = self.parent._get_attr_or_parents(attr)
        return res

    def __str__(self):
        return (self.__class__.__module__ + '.' +
                self.__class__.__name__ + ': "' + self.name + '"')

    @property
    def schemas(self):
        return self._get_attr_or_parents('_schemas')

    @property
    def specifications(self):
        return self._get_attr_or_parents('_specifications')

    @property
    def armt_definitions(self):
        return self._get_attr_or_parents('_armt_definitions')

    @property
    def armt_protocols(self):
        return self._get_attr_or_parents('_armt_protocols')


class Project(RadarObject):
    """
    """
    def __init__(self, path, **kwargs):
        """
        Parameters
        __________
        path : list / str
            Path(s) containing project data
        name : str
            The name of the project
        participants : list
            A list of participants to load in paths
        blacklist : list
            Files/folders to ignore in paths
        schemas : None
            --
        specifications : None
            --
        armt_definitions : None
            --
        armt_protocols : None
            --
        """
        super(Project, self).__init__(path, **kwargs)
        self._keys_view = ProjectKeysView
        self._items_view = ProjectItemsView
        self._values_view = ProjectValuesView
        self._subobject_class = Participant
        self._schemas = kwargs.get('schemas')
        self._specifications = kwargs.get('specifications')
        self._armt_definitions = kwargs.get('armt_definitions')
        self._armt_protocols = kwargs.get('armt_protocols')
        self._whitelist = kwargs.get('participants')
        self._blacklist = kwargs.get('blacklist', [])

    def _list_subobjects(self):
        all_paths = self.fs.listdir(self.path)
        whitelist = [p['name'] for p in all_paths] if self._whitelist is None \
            else self._whitelist
        blacklist = self._blacklist
        return [p for p in all_paths if
                p['name'] in whitelist and
                p['name'] not in blacklist]

    @property
    def participants(self):
        return self.keys()


class ProjectKeysView(KeysView):
    def __repr__(self):
        keys_str = ', '.join(self._mapping.participants)
        return 'ProjectParticipantKeys[' + keys_str + ']'


class ProjectItemsView(ItemsView):
    def __repr__(self):
        items_str = ', '.join(['("{name}": {p})'.format(name=name,
                                                        p=self._mapping[name])
                               for name in self._mapping.participants])
        return 'ProjectParticipantItems[' + items_str + ']'


class ProjectValuesView(ValuesView):
    def __repr__(self):
        vals_str = ', '.join([str(self._mapping[name])
                              for name in self._mapping.participants])
        return 'ProjectParticipantValues[' + vals_str + ']'


class Participant(RadarObject):
    """ A class to hold data and methods concerning participants/subjects in a
    RADAR trial. Typically intialised by opening a Project.
    """
    def __init__(self, path, **kwargs):
        super(Participant, self).__init__(path, **kwargs)
        self._keys_view = ParticipantKeysView
        self._items_view = ParticipantItemsView
        self._values_view = ParticipantValuesView
        self._subobject_class = RadarData

    def __repr__(self):
        return f'{self.__class__} Participant: {self.name}'

    @property
    def data(self):
        return self.keys()


class ParticipantKeysView(KeysView):
    def __repr__(self):
        keys_str = ', '.join(self._mapping.data)
        return 'ParticipantDataKeys[' + keys_str + ']'


class ParticipantItemsView(ItemsView):
    def __repr__(self):
        items_str = ', '.join(['("{name}": {p})'.format(name=name,
                                                        p=self._mapping[name])
                               for name in self._mapping.data])
        return 'ParticipantDataItems[' + items_str + ']'


class ParticipantValuesView(ValuesView):
    def __repr__(self):
        vals_str = ', '.join([str(self._mapping[name])
                              for name in self._mapping.data])
        return 'ParticipantDataValues[' + vals_str + ']'


class RadarData():
    def __repr__(self):
        rep = f'{self.name}'
        if self._files:
            start_date = fsfilename_date(self.files[0], self.fs.sep)
            end_date = fsfilename_date(self.files[-1], self.fs.sep)
            rep += f' [{len(self._files)} files]'
            rep += f'{start_date.strftime("%Y-%m-%dT%H%M")} : '
            rep += f'{end_date.strftime("%Y-%m-%dT%H%M")}'
        elif self._files is None:
            rep += ' [unlisted files]'
        else:
            rep += ' [0 files]'
        if self.schema is not None:
            rep += '\n'
            rep += '\n'.join([f'{k:<16} {"Timestamp" if k in self.schema.timecols else v.nptype}'
                              for k, v in self.schema.fields.items()])
        return rep

    @property
    def doc(self):
        print(self.name)
        for k, v in self.schema.fields.items():
            line = textwrap.wrap(f'{k:<16} {v.avro.doc}', 80)
            print(('\n' + ' ' * 17).join(line))

    def __init__(self, path, parent=None, name=None, *args, **kwargs):
        self.parent = parent
        self._files = kwargs.pop('files', None)
        specs = infer_storage_options(path)
        self.path = specs.pop('path')
        fs = kwargs.get('fs')
        if fs is None:
            fs = filesystem(**specs)
        self.fs = fs
        self.name = name if name else self.path.split(self.fs.sep)[-1]
        schema = None
        specification = kwargs.get('specification', specifications.get(self.name))
        if specification is not None:
            schema = schemas.get('org.radarcns' +
                                 specification.value_schema)
        dtypes = schema.dtypes if schema is not None else {}
        timecols = schema.timecols if schema is not None else []
        self.schema = schema
        self.reader = RadarCsvReader(dtypes=dtypes, timecols=timecols)

    def _populate_files(self):
        files = self.fs.listdir(self.path)
        files.sort(key=lambda x: x['name'])
        self._files = [f for f in files if f['type'] == 'file' and
                       (f.__setitem__('basename',
                                      f['name'].split(self.fs.sep)[-1])
                        is None) and
                       (f['name'][-6:] == 'csv.gz' or f['name'][-3:] == 'csv')]

    @property
    def files(self):
        if self._files is None:
            self._populate_files()
        return self._files

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step:
                raise KeyError('Step not supported')
            return self.between_dates(key.start, key.stop)
        else:
            raise KeyError('Only datetime slices currently supported')

    def between_dates(self, start=None, end=None):
        def meets_cond(epoch):
            return start < epoch < end

        start = float('-inf') if start is None else to_ts(start)
        end = float('+inf') if end is None else to_ts(end)
        return self._subset([f for f in self.files if
                             meets_cond(filename_to_epoch(f['basename']).timestamp())])

    def between_mtime(self, start=None, end=None):
        def meets_cond(epoch):
            return start < epoch < end
        start = float('-inf') if start is None else to_ts(start)
        end = float('+inf') if end is None else to_ts(end)
        return self._subset([f for f in self.files if
                             meets_cond(f['mtime'])])

    def _subset(self, files):
        return RadarData(self.path, self.parent, self.name, files=files, fs=self.fs)

    def to_dask_dataframe(self):
        dread = delayed(self.reader)
        divisions = [fsfilename_date(f, self.fs.sep).timestamp() for f in self.files]
        divisions.append(divisions[-1] + 3600)
        divisions = (np.array(divisions) * 1e9).astype('M8[ns]')
        return dd.from_delayed([dread(f['name'], self.fs,
                                      compression='gzip' if f['name'][-2:] == 'gz' else None)
                                for f in self.files],
                               divisions=divisions)

    def to_pandas(self):
        dataframes = [self.reader(f['name'], self.fs,
                                  compression='gzip' if f['name'][-2:] == 'gz' else None)
                      for f in self.files]
        if dataframes:
            return pd.concat(dataframes)

    def to_numpy(self):
        return self.to_pandas().reset_index().values


def filename_to_epoch(fn):
    return datetime.strptime(fn[:13], '%Y%m%d_%H%M')


def fn_from_fsspec(f, sep):
    return f['name'].split(sep)[-1]


def fsfilename_date(f, sep):
    fn = fn_from_fsspec(f, sep)
    return filename_to_epoch(fn)


def to_ts(timestamp_like):
    if isinstance(timestamp_like, float):
        if timestamp_like > 1e17:
            return timestamp_like / 1e9
        else:
            return timestamp_like
    return pd.Timestamp(timestamp_like).value / 1e9



#!/usr/bin/env python3
import json
import requests
import numpy as np
import pandas as pd
from functools import lru_cache
from ..defaults import config, protocols

class Object(object):
    pass

class Protocol(object):
    def __init__(self, json_protocol):
        self._json = json_protocol
        self.name = self._json['name']
        self.completition_time = self._json['estimatedCompletionTime']
        self.questionnaire = Object()
        self.questionnaire.__dict__ = self._json['questionnaire']
        self.questionnaire.github_owner, self.questionnaire.github_repo = \
                self.questionnaire.repository.split('.com/')[1].split('/')[0:2]
        self.questionnaire.definitions = {}
        self.questionnaire.github_file = 'questionnaires/{0}/{0}_armt.json'.\
                                          format(self.questionnaire.name)

        prot = self._json['protocol']
        self.repeat_protocol = Object()
        self.repeat_protocol.__dict__ = prot['repeatProtocol']
        self.repeat_questionnaire = Object()
        self.repeat_questionnaire.__dict__ = prot['repeatQuestionnaire']
        self.reminders = Object()
        self.reminders.__dict__ = prot['reminders']


    @lru_cache(maxsize=2)
    def _get_definition_commits(self):
        """ A function to retrieve git commits from the Protocol object repo
        Returns
        _______
        commits : pd.DataFrame
            Dataframe with a 'date' and 'sha' column for each git commit
        """
        url = 'https://api.github.com/repos/{}/{}/commits?path={}&per_page=100'\
                .format(self.questionnaire.github_owner,
                        self.questionnaire.github_repo,
                        self.questionnaire.github_file)
        r = requests.get(url)
        r.raise_for_status()
        c_json = json.loads(r.text)
        commits = pd.DataFrame(columns=['date', 'sha'],
                               data=[[c['commit']['author']['date'], c['sha']]
                                      for c in c_json])
        commits['date'] = pd.to_datetime(commits['date'])
        commits.set_index('date', inplace=True)
        return commits

    @lru_cache(maxsize=64)
    def get_definition(self, commit='master'):
        """ Retrieves the aRMT definition corresponding to a particular commit
        Parameters
        _________
        commit : str
            sha of the commit. (default: 'master')

        Returns
        ________
        definition : list of dicts
            Definition parsed by json.loads.
            Normally this should be a list containing dicts for each question.

        """
        url = 'https://raw.githubusercontent.com/{}/{}/{}/{}'\
                .format(self.questionnaire.github_owner,
                       self.questionnaire.github_repo,
                       commit,
                       self.questionnaire.github_file)
        r = requests.get(url)
        r.raise_for_status()
        definition = json.loads(r.text)
        return definition

    def get_definition_before_date(self, date):
        """ Retrieves a defintion directly prior to a given date
        Parameters
        __________
        date : str or datetime/Timestamp object
            Use this date to get the definition in use at that time.

        Returns
        ________
        definition : list of dicts
            Definition parsed by json.loads.
            Normally this should be a list containing dicts for each question.
        """
        commits = self._get_definition_commits()
        idx = commits[commits.index < date].index
        if len(idx) == 0:
            raise ValueError(('There are no definition commits for "{}" in '
                              '"{}" before the date "{}"'.format(
                                  self.questionnaire.github_file,
                                  self.questionnaire.github_repo,
                                  date)))
        sha = commits.loc[idx.max()]['sha']
        return self.get_definition(commit=sha)

    def protocol_timedelta(self):
        amount = self.repeat_protocol.amount
        unit = self.repeat_protocol.unit
        if unit == 'week':
            unit = 'W'
        elif unit == 'year':
            unit = 'W'
        return pd.Timedelta('{} {}'.format(amount, unit))

    def questionnaire_timedeltas(self):
        return pd.Series([pd.Timedelta('{} {}'\
                    .format(i, self.repeat_questionnaire.unit))
                    for i in self.repeat_questionnaire.unitsFromZero])

    def expected_questionnaires(self, time_delta, initial=True, offset='0'):
        timedeltas = pd.Series(dtype='timedelta64[ns]')
        time_delta = pd.Timedelta(time_delta)
        offset = pd.Timedelta(offset)
        repeats = np.ceil(time_delta / self.protocol_timedelta()).astype(int)
        for i in range(0 if initial else 1, repeats):
            deltas = i * self.protocol_timedelta() + \
                    self.questionnaire_timedeltas() + \
                    offset
            timedeltas = timedeltas.append(deltas)
        return pd.TimedeltaIndex(timedeltas[timedeltas < time_delta])

    def questionnaires_between_times(self, start_time, end_time,
                                     enrolment_time=None,
                                     initial=True, offset='0'):
        start_time = pd.Timestamp(start_time)
        end_time = pd.Timestamp(end_time)
        if enrolment_time:
            enrolment_time = pd.Timestamp(enrolment_time)
        else:
            enrolment_time = start_time
        time_delta = end_time - enrolment_time
        deltas = self.expected_questionnaires(time_delta=time_delta,
                                              initial=initial, offset=offset)
        times = enrolment_time + deltas
        return times[times >= start_time + pd.Timedelta(offset)]

    def adherence(self, df, start_time, end_time, **kwargs):
        """ Calculates adherence for a given aRMT dataframe
        Parameters
        _________
        df : pd.DataFrame
            A dataframe with a datetime index
            corresponding to questionnaire submission times.
        start_time : pd.Timestamp
            The start of the time frame of interest.
            Anything that pandas can convert to a Timestamp.
        end_time : pd.Timestamp
            The end of the time frame of interest.
        enrolment_time : pd.Timestamp or None (default: None)
            The time of enrolment, if different from start_time
        initial : bool (optional)
            Whether the questionnaires start at time-zero or after the first
            repeat period. (default: True)
        offset : pd.Timedelta (optional)
            An offset to apply to the start_time from which point
            questionnaires are delivered. (default: '0')
            Anything that pandas can convert to a Timedelta
        """
        def interval_sum(completed, expected, starti, endi):
            return sum(np.logical_and((completed > expected[starti]),
                                      (completed < expected[endi])))

        expected_times = self.questionnaires_between_times(
                start_time, end_time, **kwargs)
        completed_times = df.index
        exp_end = expected_times.append(pd.DatetimeIndex([end_time]))
        intervals = [interval_sum(completed_times, exp_end, i, i+1)
                     for i in range(len(expected_times))]
        return pd.DataFrame(data={'time': expected_times.tz_convert(None).values,
                                  'questions_completed': intervals})


def from_url(url):
    r = requests.get(url)
    r.raise_for_status()
    protocols = json.loads(r.text)
    protocols = {p['name']: Protocol(p) for p in protocols['protocols']}
    return protocols


def from_path(path):
    with open(path, 'r') as f:
        protocols_file = json.load(f)
    protocols = {p['name']: Protocol(p) for p in protocols_file['protocols']}
    return protocols


if config.protocol.path:
    protocols.update(from_path(config.protocol.path))
elif config.protocol.url:
    protocols.update(from_url(config.protocol.url))



#!/usr/bin/env python3
from scipy import signal


def butterworth(arr, cutoff, freq, order=5, ftype='highpass'):
    """ butterworth filters the array through a two-pass butterworth filter.
    Parameters
    __________
    arr: list or numpy.array
        The timeseries array on which the filter is applied
    cutoff: scalar or len-2 sequence
        The critical frequencies for the Butterworth filter. The point at
        which the gain drops to 1/sqrt(2) of the passband. Must be length-2
        sequence for a bandpass filter giving [low, high]. Or else the scalar
        cutoff frequency for either a low-pass' or 'highpass' filter.
    freq: float
        The frequency (Hz) of the input array
    ftype: {'highpass', 'lowpass', 'bandpass'}
        The filter type. Default is 'highpass'
    """
    nyq = 0.5 * freq
    b, a = signal.butter(order, cutoff/nyq, ftype)
    return signal.filtfilt(b, a, arr)



#!/usr/bin/env python3
import base64
from collections import Counter
import numpy as np


def convert_audio_data(x):
    return np.array(base64.b64decode(''.join(x.split('\\n')))
                    .split(b'\n')[1].split(b';')[1:], dtype=float)



""" Location plots using cartopy.
Cartopy requires several libraries not available through pip.
You should either install the requirements through your package manager
or install cartopy using conda.
https://scitools.org.uk/cartopy/docs/latest/installing.html
"""
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
try:
    import cartopy.crs as ccrs
    from cartopy.io.img_tiles import Stamen
except ModuleNotFoundError:
    raise ImportError(
        'Geo plotting requires cartopy\n'
        'Installing with conda is the most straightforward approach: \n'
        '    conda install -c conda-forge cartopy\n'
        'https://scitools.org.uk/cartopy/docs/latest/installing.html'
    )


def geo_scatter(latitudes, longitudes, projection=None,
                tiler=Stamen('toner-background'), zoom_level=8,
                extent=None, alpha=0.3, ax=None, fig=None, **kwargs):
    """ Latitude / longitude scatter plot
    Parameters:
        latitudes (float[n]): Numpy array of latitudes
        longitudes (float[n]): Numpy array of longitudes
        projection: A cartopy projection
        tiler: A cartopy map tiler / background. Default: Stamen terrain
        zoom_level (int): Tiler zoom level. Default: 8
        extent (float[4]): [Long_min, Long_max, Lat_min, Lat_max]
        ax: matplotlib axes object. Will create one if not given.
        fig: matplotlib figure. Will create one if not given
        alpha (float): Alpha value for scatter points
        kwargs: other kwargs to provide to matplotlib.scatter
    Returns:
        matplotlib.figure.Figure
    """
    if projection is None:
        projection = ccrs.PlateCarree()
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1, projection=projection)
    if extent:
        ax.set_extent(extent, crs=projection)
    if tiler:
        ax.add_image(tiler, zoom_level)
    else:
        ax.coastlines('10m')
    ax.scatter(longitudes, latitudes, alpha=alpha,
               transform=projection, **kwargs)
    scale_bar(ax)
    gl = ax.gridlines(draw_labels=True, color='white')
    gl.xlabels_top = False
    gl.ylabels_right = False
    return fig


def geo_heatmap(latitudes, longitudes, projection=None,
                tiler=Stamen('toner-background'), zoom_level=8,
                extent=None, bins=40, ax=None, fig=None, **kwargs):
    """ Latitude / longitude heatmap
    Parameters:
        latitudes (float[n]): Numpy array of latitudes
        longitudes (float[n]): Numpy array of longitudes
        projection: A cartopy projection
        tiler: A cartopy map tiler / background. Default: Stamen terrain
        zoom_level (int): Tiler zoom level. Default: 8
        extent (float[4]): [Long_min, Long_max, Lat_min, Lat_max]
        bins (int): Number of bins for the heatmap
        ax: matplotlib axes object. Will create one if not given.
        fig: matplotlib figure. Will create one if not given
        kwargs: other kwargs to provide to ax.hist2d
    Returns:
        matplotlib.figure.Figure
    """
    cmap = mpl.cm.Reds
    n = cmap.N
    cmap = cmap(np.arange(n))
    cmap[:, -1] = np.linspace(0.7, 0.99, n)
    cmap[0, -1] = 0
    cmap = mpl.colors.ListedColormap(cmap)
    if projection is None:
        projection = ccrs.PlateCarree()
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1, projection=projection)
    if not extent:
        extent = [longitudes.min()-0.1, longitudes.max()+0.1,
                  latitudes.min()-0.1, latitudes.max()+0.1]
    ax.set_extent(extent, crs=projection)
    if tiler:
        ax.add_image(tiler, zoom_level)
    else:
        ax.coastlines('10m')
    xy = ax.projection.transform_points(ccrs.Geodetic(), longitudes.values,
                                        latitudes.values)
    ax.hist2d(xy[:, 0], xy[:, 1], norm=mpl.colors.LogNorm(), cmap=cmap,
              bins=bins, zorder=10, **kwargs)
    scale_bar(ax)
    gl = ax.gridlines(draw_labels=True, color='white')
    gl.xlabels_top = False
    gl.ylabels_right = False
    return fig


def geo_hexmap(latitudes, longitudes, projection=None,
               tiler=Stamen('toner-background'), zoom_level=8,
               extent=None, gridsize=30, ax=None, fig=None, **kwargs):
    """ Latitude / longitude hexmap
    Parameters:
        latitudes (float[n]): Numpy array of latitudes
        longitudes (float[n]): Numpy array of longitudes
        projection: A cartopy projection
        tiler: A cartopy map tiler / background. Default: Stamen terrain
        zoom_level (int): Tiler zoom level. Default: 8
        extent (float[4]): [Long_min, Long_max, Lat_min, Lat_max]
        gridsize (int): Gridsize of the hexs. Default: 30
        ax: matplotlib axes object. Will create one if not given.
        fig: matplotlib figure. Will create one if not given
        kwargs: other kwargs to provide to ax.hist2d
    Returns:
        matplotlib.figure.Figure
    """
    cmap = mpl.cm.Reds
    n = cmap.N
    cmap = cmap(np.arange(n))
    cmap[:, -1] = np.linspace(0.7, 0.99, n)
    cmap[0, -1] = 0
    cmap = mpl.colors.ListedColormap(cmap)
    if projection is None:
        projection = ccrs.PlateCarree()
    if ax is None:
        if fig is None:
            fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(1, 1, 1, projection=projection)
    if not extent:
        extent = [longitudes.min()-0.05, longitudes.max()+0.05,
                  latitudes.min()-0.05, latitudes.max()+0.05]
    ax.set_extent(extent, crs=projection)
    if tiler:
        ax.add_image(tiler, zoom_level)
    else:
        ax.coastlines('10m')
    xy = ax.projection.transform_points(ccrs.Geodetic(), longitudes.values,
                                        latitudes.values)
    ax.hexbin(xy[:, 0], xy[:, 1], norm=mpl.colors.LogNorm(), cmap=cmap,
              gridsize=gridsize, zorder=10, edgecolors=None, **kwargs)
    scale_bar(ax)
    gl = ax.gridlines(draw_labels=True, color='white')
    gl.xlabels_top = False
    gl.ylabels_right = False
    return fig


def scale_bar(ax, length=None, location=(0.5, 0.05), linewidth=3):
    """ Scale bar for cartopy plots
    Parameters:
        ax: matplotlib axes to draw on.
        length (int): length of scalebar in km
        location (tuple(float, float)): is center of the scalebar
        linewidth (float): Thickness of the scalebar.
    Returns:
        None
    """
    llx0, llx1, lly0, lly1 = ax.get_extent(ccrs.PlateCarree())
    sbllx = (llx1 + llx0) / 2
    sblly = lly0 + (lly1 - lly0) * location[1]
    tmc = ccrs.TransverseMercator(sbllx, sblly)
    x0, x1, y0, y1 = ax.get_extent(tmc)
    sbx = x0 + (x1 - x0) * location[0]
    sby = y0 + (y1 - y0) * location[1]

    def scale_number(x):
        if str(x)[0] in ['1', '2', '5']:
            return int(x)
        return scale_number(x - 10 ** ndim)
    if not length:
        length = (x1 - x0) / 5000
        ndim = int(np.floor(np.log10(length)))
        length = round(length, -ndim)
        length = scale_number(length)

    bar_xs = [sbx - length * 500, sbx + length * 500]
    ax.plot(bar_xs, [sby, sby], transform=tmc, color='k', linewidth=linewidth)
    ax.text(sbx, sby, str(length) + ' km', transform=tmc,
            horizontalalignment='center', verticalalignment='bottom')
